#include <stdio.h>
#include <stdlib.h> /* calloc() */
#include <fcntl.h>  /* open()   */
#include <unistd.h> /* read()   */

#define NUM_COUNTERS 2048
#define COUNT_SHIFT  2
#define COUNT_MASK   0x7ff

int *counters;

static int predict(unsigned int addr) {
   
   register int line;
   
   line=(addr>>COUNT_SHIFT)&COUNT_MASK;

   if (counters[line]>=2) return 1;
   
   return 0;
}

static void update_counters(unsigned int addr,int was_taken) {
   
   register int line;
   
   line=(addr>>COUNT_SHIFT)&COUNT_MASK;
   
   if (was_taken) {
      if (counters[line]<3) counters[line]++;
   }
   else  {
      if (counters[line]>0) counters[line]--;
   }
}

#define NOT_A_BRANCH  0
#define COND_BRANCH   1
#define UNCOND_BRANCH 2
#define LIKELY_BRANCH 3  /* p32 of r12k manual - likely don't update predictor*/

static int get_branch_type(unsigned int insn) {
 
   unsigned int opcode,func,rs,rt;
   
   opcode=(insn>>26)&0x3f;
   func=(insn&0x3f);
   rs=(insn>>21)&0x1f;
   rt=(insn>>16)&0x1f;
   
      /* beq */
   if (opcode==0x4) {   
      if ((rs==0)&&(rt==0)) {
	 return UNCOND_BRANCH; /* b */
      
      } else {
         return COND_BRANCH; /* beq */
      }
   }
   
      /* bgezal */
   if ((opcode==0x1) && (rt==0x11)) {
      if (rs==0) return UNCOND_BRANCH; /* bal */
      else return COND_BRANCH; /* begzal */
   }
   
   if ((opcode==0x11) && (rs==0x8)) return COND_BRANCH; /* bc1f  */
//   if ((opcode==0x12) && (rs==8)) return COND_BRANCH; /* bc2f  */ no c2 ever

   if (opcode==0x14) return LIKELY_BRANCH; /* beql */
   if ((opcode==0x1) && (rt==0x1)) return COND_BRANCH; /* bgez */
   if ((opcode==0x1) && (rt==0x13)) return LIKELY_BRANCH; /* bgezall */
   if ((opcode==0x1) && (rt==0x3)) return LIKELY_BRANCH; /* bgezl */
   if ((opcode==0x7)) return COND_BRANCH; /* bgtz */
   if ((opcode==0x17)) return LIKELY_BRANCH; /* bgtzl */
   if ((opcode==0x6)) return COND_BRANCH; /* blez */
   if ((opcode==0x16)) return LIKELY_BRANCH; /* blezl */   
   if ((opcode==0x1) && (rt==0x0)) return COND_BRANCH; /* bltz */
   if ((opcode==0x1) && (rt==0x10)) return COND_BRANCH; /* bltzal */
   if ((opcode==0x1) && (rt==0x12)) return LIKELY_BRANCH; /* bltzall */
   if ((opcode==0x1) && (rt==0x2)) return LIKELY_BRANCH; /* bltzl */
   if (opcode==0x5) return COND_BRANCH; /* bne */
   if (opcode==0x15) return LIKELY_BRANCH; /* bnel */
   if (opcode==0x2) return UNCOND_BRANCH; /* j */   
   if (opcode==0x3) return UNCOND_BRANCH; /* jal */   
   if ((opcode==0x0) && (func==0x9)) return UNCOND_BRANCH; /* jalr */
   if ((opcode==0x0) && (func==0x8)) return UNCOND_BRANCH; /* jr */
   
   return NOT_A_BRANCH;

}

struct {
  unsigned int pc;
  int type;
  int taken;
} branch[3];

int main(int argc, char **argv) {
 
   int i,fd,branch_type;
   
   unsigned int buffer[1024],max_count=0,pc=0;
   
   unsigned long long cond_branches=0,uncond_branches=0,instructions=0;
   unsigned long long mispredicts=0,likely_branches=0,likely_misses=0;

   /* init counters */
   counters=calloc(NUM_COUNTERS,sizeof(int));
   for(i=0;i<NUM_COUNTERS;i++) {
      counters[i]=2;
   }
   
   fd=open("trace.bpred",O_RDONLY);
   if (fd<0) {
      fprintf(stderr,"Cannot open trace.bpred!\n");
      exit(-1);
   }
   
   while(1) {
      max_count=read(fd,buffer,4096);
      if (max_count<1) break;
      max_count/=8;
      
      instructions+=max_count;
      
      for(i=0;i<max_count;i++) {
         branch_type=get_branch_type(buffer[(i*2)+1]);
	 	 
         if (branch_type==COND_BRANCH) {
	    cond_branches++;
	 }
         else if (branch_type==UNCOND_BRANCH) {
	    uncond_branches++;
	 } 
	 else if (branch_type==LIKELY_BRANCH) { 
	    likely_branches++;
	 }
	 
	 pc=buffer[i*2];
	 
	 
	 if (branch[1].type==LIKELY_BRANCH) {
	       /* branch was not taken */
	    if (branch[1].pc+8==pc) {
	       likely_misses++;
	    }			   
	 }
	 
	    /* update counters */
	 if (branch[1].type==COND_BRANCH) {
	    
	    if (branch[1].pc+8==pc) {
	       /* branch was not taken */	       
	       update_counters(branch[1].pc,0);
	       if (branch[1].taken) {
		  mispredicts++;
	       }
	    }
	    else {
	       /* branch was taken */
	       update_counters(branch[1].pc,1);
	       if (!branch[1].taken) {
		  mispredicts++;
	       }
	    }
	 }
	    
	 branch[1].type=branch[0].type; branch[1].pc=branch[0].pc;
	 branch[1].taken=branch[0].taken;
	 
	 branch[0].type=branch_type; branch[0].pc=pc;
	 if (branch_type==COND_BRANCH) branch[0].taken=predict(pc);	 
      }	 
   }
   
   printf("Total instructions: %lld\n",instructions);
   printf("Cond   branches:    %lld\n",cond_branches);
   printf("Uncond branches:    %lld\n",uncond_branches);
   printf("Likely branches:    %lld\n",likely_branches);
   printf("Total misspredicts: %lld\n",mispredicts);
   printf("Likely misspredict: %lld\n",likely_misses);
   printf("Percentage branches:%.2f\n",((double)cond_branches+
					(double)uncond_branches+
					(double)likely_branches)*100.0/
	                                (double)instructions);
   printf("Miss rate:          %.2f\n",(double)mispredicts*100.0/(double)cond_branches);
   return 0;
}
